dofile_once("mods/koira/files/scripts/utilities.lua")
dofile_once("mods/koira/files/scripts/globals.lua")
dofile_once("data/scripts/perks/perk.lua")

function player_movement(player, is_enabled)
  local itemComponent = EntityGetFirstComponentIncludingDisabled(player, "ItemPickUpperComponent")
  local platComponent = EntityGetFirstComponentIncludingDisabled(player, "CharacterPlatformingComponent")
  local ctrlComponent = EntityGetFirstComponentIncludingDisabled(player, "ControlsComponent")

  EntitySetComponentIsEnabled(player, itemComponent, is_enabled)
  EntitySetComponentIsEnabled(player, platComponent, is_enabled)
  ComponentSetValue2(ctrlComponent, "enabled", is_enabled)

  if not is_enabled then
    local dataComp = EntityGetFirstComponent(player, "CharacterDataComponent")
    ComponentSetValueVector2(dataComp, "mVelocity", 0, 0)
  end
end

function player_visibility(player, visible)
  local sprites = EntityGetComponentIncludingDisabled(player, "SpriteComponent")
  for _, sprite in ipairs(sprites) do
    ComponentSetValue2(sprite, "visible", visible)
  end

  for _, child in ipairs(EntityGetAllChildren(player)) do
    local sprites = EntityGetComponentIncludingDisabled(child, "SpriteComponent")
    if sprites then
      for _, sprite in ipairs(sprites) do
        ComponentSetValue2(sprite, "visible", visible)
      end
    end
  end
end

local function entity_and_children_set_components_with_tag_enabled(entity_id, tag, enabled)
  EntitySetComponentsWithTagEnabled(entity_id, tag, enabled)
  for i, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
    entity_and_children_set_components_with_tag_enabled(child, tag, enabled)
  end
end

function get_active_item()
	local player = EntityGetWithTag("player_unit")[1]
	local inv = EntityGetFirstComponentIncludingDisabled(player, "Inventory2Component")
	return ComponentGetValue2(inv, "mActiveItem")   
end

function wake()
	local player = EntityGetWithTag("player_unit")[1]
	
  -- Show player again
  player_visibility(player, true)

  -- Re-enable movement
  player_movement(player, true)
  
  --show items
    local active_item = get_active_item()
    if active_item > 0 then
      entity_and_children_set_components_with_tag_enabled(active_item, "enabled_in_hand", true)
    end
	
  local endless_nightmares = GlobalsGetValue("koira_endless_nightmares", "0") == "1"
  local x, y = LOCATION(NightmareTP1)
  if endless_nightmares then
	x, y = LOCATION(NightmareTP2)
  end
  teleport_player_safely(player, x, y)
  
  local ritual_complete = GlobalsGetValue("koira_ritual_complete", "0") == "1"
  if not ritual_complete then
	local STARTING_PERKS = {
	  "DOUBLE_JUMP",
	  "DASH",
	}
	
	function enable_perks(player, perks)
	  local x, y = EntityGetTransform(player)
	  for _,name in ipairs(perks) do
		local perk = perk_spawn(x, y, name)
		if ( perk ~= nil ) then
		  perk_pickup(perk, player, EntityGetName(perk), false, false)
		end
	  end
	end

    enable_perks_without_icons(player, STARTING_PERKS)
	
	GamePrintImportant( GameTextGetTranslatedOrNot("$koira_dream_enter1"), GameTextGetTranslatedOrNot("$koira_dream_enter2") )
  end

  -- Once we get out of nightmares, we are welcome to interact with the altar
  GlobalsSetValue(ALTAR_STATE, AfterNightmares)

  -- No one is friendly in your dreams
  EntitySetValue(player, "GenomeDataComponent", "herd_id", StringToHerdId("player"))
end

-- function bed_empty()
  -- --play sleeping animation
  -- local sprite_component = EntityGetFirstComponentIncludingDisabled(GetUpdatedEntityID(), "SpriteComponent")
  
  -- ComponentSetValue2(sprite_component, "rect_animation", "stand")
  -- ComponentSetValue2(sprite_component, "next_rect_animation", "stand")
  -- print("bed turned empty")
-- end

function interacting(player, bed, bed_name)
  local var = tonumber(GlobalsGetValue("koira_sleeping_complete_times", "0"))
  GlobalsSetValue("koira_sleeping_complete_times", var + 1)
  GlobalsSetValue("koira_is_sleeping", "1")
  
  -- 1. Stop player from moving by outside forces or controls
  player_movement(player, false)

  -- 2. Make player invisible
  player_visibility(player, false)
  
  --hide items
    local active_item = get_active_item()
    if active_item > 0 then
      entity_and_children_set_components_with_tag_enabled(active_item, "enabled_in_hand", false)
    end
  
	function get_child_with_name(entity_id, name)
	  for i, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
		if EntityGetName(child) == name then
		  return child
		end
	  end
	end
  
  local sleep = get_child_with_name(player, "sleep")
	if sleep then
	  EntityKill(sleep)
	end
  
  GamePrint(GameTextGetTranslatedOrNot("$koira_dream_enter_prompt"))
  
  --play sleeping animation
  local sprite_component = EntityGetFirstComponentIncludingDisabled(bed, "SpriteComponent")
  
  ComponentSetValue2(sprite_component, "rect_animation", "sleep")
  ComponentSetValue2(sprite_component, "next_rect_animation", "stand")
  
  -- SetTimeOut(4, "mods/koira/files/biomes/objects/bed/sleep.lua", "bed_empty")
  SetTimeOut(4, "mods/koira/files/biomes/objects/bed/sleep.lua", "wake")
  
  local entity_id = GetUpdatedEntityID()
  local interacting = EntityGetFirstComponent(entity_id, "InteractableComponent")
  EntitySetComponentIsEnabled( entity_id, interacting, false )

end
