dofile_once("data/scripts/lib/utilities.lua")
dofile_once("mods/koira/files/scripts/utilities.lua")
dofile_once("mods/koira/files/scripts/globals.lua")
-- dofile_once("mods/koira/files/coroutines.lua")
dofile_once("mods/koira/files/util.lua")
dofile_once("mods/koira/files/camera.lua")
dofile_once("mods/koira/files/scripts/gun_utilities.lua")


local dog = GetUpdatedEntityID()
local dog_x, dog_y, dog_rot, dog_scale_x = EntityGetTransform(dog)

local function entity_and_children_set_components_with_tag_enabled(entity_id, tag, enabled)
  EntitySetComponentsWithTagEnabled(entity_id, tag, enabled)
  for i, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
    entity_and_children_set_components_with_tag_enabled(child, tag, enabled)
  end
end

function get_active_item()
	local player = EntityGetWithTag("player_unit")[1]
	local inv = EntityGetFirstComponentIncludingDisabled(player, "Inventory2Component")
	return ComponentGetValue2(inv, "mActiveItem")   
end

function get_state()
  local entity_id = GetUpdatedEntityID()
  state = state or {}
  state[entity_id] = state[entity_id] or {}
  return state[entity_id]
end

function SetValueBool(key, value)
  get_state()[key] = not not value
end

function GetValueBool(key, default_value)
  local val = get_state()[key]
  if val ~= nil then
    return val
  else
    return not not default_value
  end
end

local bone = EntityGetWithName("the_bone")


-- Trigger for the_bone's LuaComponent
function throw_item(from_x, from_y, to_x, to_y)
  GamePrint("$koira_dog1")
  DOG_STATE = FollowBone
end
---


function dog_goto(x, y)
  local player = get_player()
  local px, py, prot, pscale_x = EntityGetTransform(get_player())
  
  local vel_x, vel_y = rad_to_vec(get_direction(dog_x, dog_y, x, y))
  
  local ending_triggered = GetValueBool("ending", false)
  local dog_in_position = GetValueBool("dog_in_position", false)
  local sleeping = GlobalsGetValue("koira_is_sleeping", "0") == "1"
  
  if ending_triggered and ( px <= dog_x + 11 and px >= dog_x - 11 ) then
	  if not dog_in_position and not sleeping then
		EntityAddComponent(get_player(), "LuaComponent", {
			script_source_file="mods/koira/files/enemies/the_dog/roll_credits.lua",
			execute_every_n_frame="1",
			execute_on_added="1"
		} )
		
		EntitySetComponentsWithTagEnabled(dog, "love", true)
		local dog_sprite_component = EntityGetFirstComponent(dog, "SpriteComponent")
		local ritual_complete = GlobalsGetValue("koira_ritual_complete", "0") == "1"
		if not ritual_complete then
			ComponentRemoveTag(dog_sprite_component, "character" )
			ComponentSetValue2(dog_sprite_component, "rect_animation", "happy")
			ComponentSetValue2(dog_sprite_component, "next_rect_animation", "happy")
		else
			ComponentRemoveTag(dog_sprite_component, "character" )
			ComponentSetValue2(dog_sprite_component, "rect_animation", "lie_down")
			ComponentSetValue2(dog_sprite_component, "next_rect_animation", "lie_down")
			
			local sprite_components = EntityGetComponent(player, "SpriteComponent") or {}
			
			for i, sprite_component in ipairs(sprite_components) do

				if ComponentGetValue2(sprite_component, "image_file") == "data/enemies_gfx/player.xml"then
					ComponentRemoveTag(sprite_component, "character" )
					ComponentSetValue2(sprite_component, "rect_animation", "eat")
					ComponentSetValue2(sprite_component, "next_rect_animation", "eat")
					
				end
				
			end
		end
		
		if pscale_x > 0 and px < dog_x then
			EntitySetTransform(dog, dog_x, dog_y, dog_rot, -1)
		elseif pscale_x < 0 and px > dog_x then
			EntitySetTransform(dog, dog_x, dog_y, dog_rot, 1)
		end
		
		SetValueBool( "dog_in_position", true )
	  end
  else
	  local dog_data = EntityGetFirstComponent(dog, "CharacterDataComponent")
	local vel_x2, vel_y2 = ComponentGetValue2( dog_data, "mVelocity" )
	if not sleeping then
	  ComponentSetValueVector2(dog_data, "mVelocity", vel_x*-200, vel_y2)
	else
	  local sprite_comp = EntityGetFirstComponentIncludingDisabled(dog,"SpriteComponent")
	  local alpha = ComponentGetValue2(sprite_comp, "alpha")
	  ComponentSetValue2(sprite_comp, "alpha", alpha - alpha * 0.01)
	end
  end
end


function dog_pickup_bone(bone)
  EntitySetComponentsWithTagEnabled(dog, "fake_bone", true)
  EntityKill(bone)
end


function dog_drop_bone(bone)
  EntitySetComponentsWithTagEnabled(dog, "fake_bone", false)
  EntityLoad("mods/koira/files/items/the_bone/the_bone.xml", dog_x-12, dog_y)
end


function main()
	-- print("Is coroutine running?: " .. tostring(coroutine.running()))
  local dog_in_position = GetValueBool("dog_in_position", false)
  local sleeping = GlobalsGetValue("koira_is_sleeping", "0") == "1"
  if not dog_in_position and get_player() > 0 then
	local enemies = EntityGetInRadiusWithTag(dog_x, dog_y, 250, "enemy")
	local px, py, prot, pscale_x = EntityGetTransform(get_player())
	
	if #enemies == 0 then
		
		local p_distance = get_distance(dog_x, dog_y, px, py)
		if p_distance < 150 then
			SetValueBool( "ending", true )
			local camera_fixer_added = GetValueBool("camer_fixer_added", false)
			
			if not camera_fixer_added and not sleeping then
				if dog_scale_x > 0 and dog_x < px then
					EntitySetTransform(get_player(), px, py, prot, -1)
				elseif dog_scale_x < 0 and dog_x > px then
					EntitySetTransform(get_player(), px, py, prot, 1)
				end
				
				GlobalsSetValue("koira_ending_triggered", "1")
				
			    set_controls_enabled(false)
				GameSetCameraFree(true)
				local camera_fixer = EntityCreateNew()
				EntitySetTransform(camera_fixer, px, py)
				EntityAddChild(get_player(), camera_fixer)
				EntityAddComponent(camera_fixer, "InheritTransformComponent", {} )
				EntityAddComponent(camera_fixer, "LuaComponent", {
					script_source_file="mods/koira/files/enemies/the_dog/camera_fixer.lua",
					execute_every_n_frame="1",
					execute_on_added="1"
				} )

			  --hide items
				local active_item = get_active_item()
				if active_item > 0 then
				  entity_and_children_set_components_with_tag_enabled(active_item, "enabled_in_hand", false)
				end
				SetValueBool( "camer_fixer_added", true )
				
				local damage_model_component = EntityGetFirstComponentIncludingDisabled(get_player(), "DamageModelComponent")
				ComponentSetValue2(damage_model_component, "invincibility_frames", 999999999999)
			end
			
			if ending_triggered then
				local controls_component = EntityGetFirstComponentIncludingDisabled(get_player(), "ControlsComponent")
				if pscale_x > 0 then
					ComponentSetValue2(controls_component, "mAimingVector", 1, 0)
					ComponentSetValue2(controls_component, "mAimingVectorNormalized", 5000, 0)
					ComponentSetValue2(controls_component, "mAimingVectorNonZeroLatest", 5000, 0)
					ComponentSetValue2(controls_component, "mMousePosition", 5000, 0)
				elseif pscale_x < 0 then
					ComponentSetValue2(controls_component, "mAimingVector", 1, 0)
					ComponentSetValue2(controls_component, "mAimingVectorNormalized", -5000, 0)
					ComponentSetValue2(controls_component, "mAimingVectorNonZeroLatest", -5000, 0)
					ComponentSetValue2(controls_component, "mMousePosition", -5000, 0)
				end
			end
			
			if pscale_x > 0 then
				dog_goto(px + 10, py)
			elseif pscale_x < 0 then
				dog_goto(px - 10, py)
			end
		end
	end
  end
end


--GamePrint("Dog state is: ".. DOG_STATE)

main()
