dofile_once("mods/koira/files/scripts/utilities.lua")
dofile_once("mods/koira/files/scripts/globals.lua")

function damage_received(damage, message, entity_thats_responsible, is_fatal)
  if is_fatal then
    if (GlobalsGetValue("DEATH_IN_PROCESS") == "1") then return end
    GlobalsSetValue("DEATH_IN_PROCESS", "1")

    fake_death()

    -- TODO: Detect if we're coming out of a sleep here
    EntitySetValue(player, "GenomeDataComponent", "herd_id", StringToHerdId("healer"))
  end
end

local function entity_and_children_set_components_with_tag_enabled(entity_id, tag, enabled)
  EntitySetComponentsWithTagEnabled(entity_id, tag, enabled)
  for i, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
    entity_and_children_set_components_with_tag_enabled(child, tag, enabled)
  end
end

function get_active_item()
	local player = EntityGetWithTag("player_unit")[1]
	local inv = EntityGetFirstComponentIncludingDisabled(player, "Inventory2Component")
	return ComponentGetValue2(inv, "mActiveItem")   
end


function fake_death()
  local player = GetUpdatedEntityID()

  -- 1. Stop player from moving by outside forces or controls
  player_movement(player, false)

  -- 2. Make player invisible
  player_visibility(player, false)

  -- 3. Make player a Hiisi again, so the enemies don't keep attacking his corpse
  entity_set_genome(player, "orcs")

  -- 3. Drop all guns here
  -- The special bottle of hatred needs to be changed/stopped
  -- GameDropAllItems(player)  -- Drops guns but not ammo?
  --GameDropPlayerInventoryItems( player )  -- Drops ammo but not guns?
  --drop_items()

  -- 4. Spawn a ragdoll in its stead..?
  local x, y = EntityGetTransform(player)
  local dummy = EntityLoad("mods/koira/files/player_ragdoll.xml", x, y)
  EntityInflictDamage(dummy, 1, "DAMAGE_CURSE", "a peasant sacrificed in your stead", "NORMAL", 5, 5, dummy, x, y, 0)
  
  --hide items
    local active_item = get_active_item()
    if active_item > 0 then
      entity_and_children_set_components_with_tag_enabled(active_item, "enabled_in_hand", false)
    end
	
	--reduce a life if not sleeping
	local lives = tonumber(GlobalsGetValue("koira_lives", "0"))
	
	local sleeping = GlobalsGetValue("koira_is_sleeping", "0") == "1"
	if not sleeping then
		GlobalsSetValue("koira_lives", lives - 1)
	end
	
	--Actually kill the player if no lives left
	if lives <= 1 then 
		GamePlaySound("mods/koira/files/audio/AdventureMode.bank", "death_sound", x, y)
		EntityKill(player)
		return
	end
	
  -- Respawn after a few seconds
  SetTimeOut(5, "mods/koira/files/scripts/death.lua", "respawn")
end

function respawn()
  
  local sleeping = GlobalsGetValue("koira_is_sleeping", "0") == "1"
  
  -- NOTE: Can't use GetUpdatedEntityID() here
  local player = get_player()
  local dmgComponent = EntityGetFirstComponentIncludingDisabled(player, "DamageModelComponent")

  local max_health = tonumber(ComponentGetValue(dmgComponent, "max_hp"))
  
  local tx, ty = LOCATION(GlobalsGetValue(CURRENT_SPAWN))
  if sleeping then
	teleport_player_safely(player, 1365, 45)
  else
	teleport_player_safely(player, tx, ty)
  end
  
	local dataComp = EntityGetFirstComponent(player, "CharacterDataComponent")
	local vel_x, vel_y = ComponentGetValue2( dataComp, "mVelocity" )
	ComponentSetValueVector2(dataComp, "mVelocity", 0, 0)
	
	-- LoadGameEffectEntityTo( player, "data/entities/misc/effect_knockback_immunity.xml" )
  
  -- local bed = EntityGetWithTag("bed")[1]
  -- local sprite_component = EntityGetFirstComponentIncludingDisabled(bed, "SpriteComponent")
  
  -- ComponentSetValue2(sprite_component, "rect_animation", "stand")
  
  ComponentSetValue2(dmgComponent, "hp", max_health)

  ---- Back from hiisi
  entity_set_genome(player, "player")

  -- Show player again
  player_visibility(player, true)

  -- Re-enable movement
  player_movement(player, true)

  -- Make sure anger returns to player
  local anger = EntityGetWithName("BOTTLED_FEELINGS")
  GamePickUpInventoryItem(player, anger)
  
  --show items
    local active_item = get_active_item()
    if active_item > 0 then
      entity_and_children_set_components_with_tag_enabled(active_item, "enabled_in_hand", true)
    end

	if sleeping then
		local ritual_complete = GlobalsGetValue("koira_ritual_complete", "0") == "1"
		if not ritual_complete then
			GameRemoveFlagRun( "PERK_PICKED_DOUBLE_JUMP" )
			GameRemoveFlagRun( "PERK_PICKED_DASH" )
		end
		
		local sleeping_complete_once = GlobalsGetValue("koira_sleeping_complete_times", "0") == "1"
		if sleeping_complete_once then
			EntityLoad("mods/koira/files/cutscene/cutscene.xml", x, y)
		end
		GlobalsSetValue("koira_is_sleeping", "0")
	end

  -- The End
  GlobalsSetValue("DEATH_IN_PROCESS", "0")
  GamePrintImportant("$koira_respawn", "")
end

function player_movement(player, is_enabled)
  local itemComponent = EntityGetFirstComponentIncludingDisabled(player, "ItemPickUpperComponent")
  local platComponent = EntityGetFirstComponentIncludingDisabled(player, "CharacterPlatformingComponent")
  local ctrlComponent = EntityGetFirstComponentIncludingDisabled(player, "ControlsComponent")

  EntitySetComponentIsEnabled(player, itemComponent, is_enabled)
  EntitySetComponentIsEnabled(player, platComponent, is_enabled)
  ComponentSetValue2(ctrlComponent, "enabled", is_enabled)

  if not is_enabled then
    local dataComp = EntityGetFirstComponent(player, "CharacterDataComponent")
    ComponentSetValueVector2(dataComp, "mVelocity", 0, 0)
  end
end

function player_visibility(player, visible)
  local sprites = EntityGetComponentIncludingDisabled(player, "SpriteComponent")
  for _, sprite in ipairs(sprites) do
    ComponentSetValue2(sprite, "visible", visible)
  end

  for _, child in ipairs(EntityGetAllChildren(player)) do
    local sprites = EntityGetComponentIncludingDisabled(child, "SpriteComponent")
    if sprites then
      for _, sprite in ipairs(sprites) do
        ComponentSetValue2(sprite, "visible", visible)
      end
    end
  end
end

function drop_items(player)
  local x, y = EntityGetTransform(player)

  local full_items = EntityGetAllChildren(EntityGetWithName("inventory_full"))
  local quick_items = EntityGetAllChildren(EntityGetWithName("inventory_quick"))

  if full_items then
    for _, item in ipairs(full_items) do
      EntityRemoveFromParent(item)
      EntitySetTransform(item, x, y)
    end
  end
  if quick_items then
    for _, item in ipairs(quick_items) do
      EntityRemoveFromParent(item)
      EntitySetTransform(item, x, y)
    end
  end
end
