dofile_once("mods/koira/files/scripts/utilities.lua")

-- Include this file with:
-- dofile_once("mods/koira/files/scripts/gun_utilities.lua")
---------------------------
-- Utilities related to our very specific gun system

--
-- General utilities
--
GUNS = {
  "knife",
  "spear",
  "pistol",
  "revolver",
  "smg",
  "shotgun",
  "assault_rifle",
  "rifle",
  "rocket_launcher",
  "grenade_launcher",
  "flammenwerfer",
  "lightning_gun",
  "railgun",
  "plasmagun",
  "stakegun",
}


function gun_entity_path(name)
  return MOD_PATH .. "weapons/" .. name .. "/gun.xml"
end


function crate_entity_path(name)
  return MOD_PATH .. "items/weapon_crates/" .. name .. ".xml"
end


--
-- Spawning functions
--
function spawn_gun(name, x, y)
  -- TODO add check for `name in GUNS`
  EntityLoad(gun_entity_path(name), x, y)
end


function spawn_all_guns(x, y)
  -- Spawn next to each other, in a nice row.
  for _, name in ipairs(GUNS) do
    spawn_gun(name, x, y)
    x = x + 20
  end
end


function spawn_crate(name, x, y)
  -- TODO add check for `name in CRATES`
  EntityLoad(crate_entity_path(name), x, y)
end


function spawn_random_weapons_crate(x, y)
  SetRandomSeed(x, y)
  local roll = Random(1, 100)

  if roll <= 30 then
    EntityLoad(crate_entity_path("small"), x, y)
  elseif roll <= 65 then
    EntityLoad(crate_entity_path("medium"), x, y)
  elseif roll <= 85 then
    EntityLoad(crate_entity_path("big"), x, y)
  elseif roll <= 100 then
    EntityLoad(crate_entity_path("energy"), x, y)
  end
end


function give_player_gun(name)
  local inv_quick = EntityGetWithName("inventory_quick")
  local gun = EntityLoad(gun_entity_path(name))
  if gun then
    EntityAddChild(inv_quick, gun)
  else
    GamePrint("Couldn't load the gun ["..name.."]")
  end
end


function give_player_random_guns(player, inventory, gun_count)
  if (inventory == nil) then return end

  local gun_randomizer = UniqueRandom()
  for i=1,gun_count do
    local gun_name = GUNS[gun_randomizer(1, #GUNS)]
    give_player_gun(player, inventory, gun_name)
  end
end


--
-- Magazine and gun stats
--
function get_gun_count()
  local items = EntityGetAllChildren(EntityGetWithName("inventory_quick"))
  if not items then
    return 0
  end

  local count = 0
  for _, item in ipairs(items) do
    if EntityHasTag(item, "GUN") then
      count = count + 1
    end
  end

  return count
end


function get_bullets_left(magazine)
  return ComponentGetValue2(
    EntityGetFirstComponentIncludingDisabled(magazine, "ItemComponent"), "uses_remaining"
  )
end


function get_action_id(magazine)
  return ComponentGetValue2(
    EntityGetFirstComponentIncludingDisabled(magazine, "ItemActionComponent"), "action_id"
  )
end


function get_mags_left(gun)
  local mags = EntityGetAllChildren(gun)
  if mags ~= nil then
    return #mags
  end
  return 0
end


function get_bandolier_size(gun)
  return ComponentObjectGetValue2(
    EntityGetFirstComponentIncludingDisabled(gun, "AbilityComponent"), "gun_config", "deck_capacity"
  )
end


function get_active_magazine(gun)
  -- Our "active" mag is always the one with least bullets. Sorry, nothing fancier.
  -- TODO: See if this is causing any performance issues.
  local mags = EntityGetAllChildren(gun)
  if mags == nil then
    return nil
  end

  table.sort(mags, function(a,b) return get_bullets_left(a) < get_bullets_left(b) end )
  return mags[1]
end


--
-- HUD controls
--
function _hud_display(enabled, entity_name, entity_path, x, y)
  local entity = EntityGetWithName(entity_name)

  if enabled then
    if entity ~= 0 then return end
    EntityLoad(entity_path, x, y)
  else
    if entity == 0 then return end
    EntityKill(entity)
  end
end


function ammo_warning_display(enabled)
  _hud_display(enabled, "low_ammo_warning", "mods/koira/files/hud/low_ammo.xml", 565, 550)
end


function mags_warning_display(enabled)
  _hud_display(enabled, "low_mags_warning", "mods/koira/files/hud/low_mags.xml", 565, 590)
end


function gun_empty_warning_display(enabled)
  _hud_display(enabled, "gun_empty_warning", "mods/koira/files/hud/gun_empty.xml", 565, 590)
end
