dofile_once("data/scripts/lib/utilities.lua")

dofile_once("mods/koira/files/scripts/utilities.lua")
dofile_once("mods/koira/files/scripts/globals.lua")

-- The implementations of our custom features
dofile_once("mods/koira/files/scripts/dash.lua")
dofile_once("mods/koira/files/scripts/death.lua")
dofile_once("mods/koira/files/scripts/jump.lua")
dofile_once("mods/koira/files/scripts/laser_sight.lua")
dofile_once("mods/koira/files/scripts/weapon_throw.lua")



-- Global stuff most of our custom scripts use
local player = GetUpdatedEntityID()
-- local px, py = EntityGetTransform(player)
-- if BiomeMapGetName( px, py ) == "THE ENDLESS NIGHTMARES" then
	-- GlobalsSetValue("koira_endless_nightmares", "1")
-- end

local activeGun = get_active_gun(player)

local ctrlComponent = EntityGetFirstComponentIncludingDisabled(player, "ControlsComponent")
local dataComponent = EntityGetFirstComponentIncludingDisabled(player, "CharacterDataComponent")
local platComponent = EntityGetFirstComponentIncludingDisabled(player, "CharacterPlatformingComponent")


-- Controls handling
local holdingUp = ComponentGetValue2(ctrlComponent, "mButtonDownUp")
local holdingRight = ComponentGetValue2(ctrlComponent, "mButtonDownRight")
local holdingLeft = ComponentGetValue2(ctrlComponent, "mButtonDownLeft")
local holdingKick = ComponentGetValue2(ctrlComponent, "mButtonDownKick")
local holdingUse = ComponentGetValue2(ctrlComponent, "mButtonDownInteract")
local holdingRmb = ComponentGetValue2(ctrlComponent, "mButtonDownRightClick")


-- Control events
local rmbClicked = (holdingRmb == true and PREV_RMB == false)

local kickClicked = (holdingKick == true and PREV_KICK == false)

local leftReleased = (holdingLeft == false and PREV_LEFT == true)
local leftClicked = (holdingLeft == true and PREV_LEFT == false)
local rightReleased = (holdingRight == false and PREV_RIGHT == true)
local rightClicked = (holdingRight == true and PREV_RIGHT == false)

local useClicked = (holdingUse == true and PREV_USE == false)
local useReleased = (holdingUse == false and PREV_USE == true)


-- Action events
local framesInAir = ComponentGetValue2(platComponent, "mFramesInAirCounter")
local framesSwimming = ComponentGetValue2(platComponent, "mFramesSwimming")

local inAir = (framesInAir > 3)
local justJumped = (inAir and PREV_FRAMES_IN_AIR <= 3)
local justLanded = (framesInAir == 0 and PREV_FRAMES_IN_AIR > 0)
local belowScalingThreshold = (framesInAir < 12)

local swimmingStarted = (framesSwimming > 0 and PREV_FRAMES_SWIMMING == 0)
local swimmingStopped = (framesSwimming == 0 and PREV_FRAMES_SWIMMING > 0)

local doGroundDash = kickClicked and (holdingLeft or holdingRight)

local isMoving = (holdingLeft or holdingUp or holdingRight or inAir)
local enableKick = (leftReleased or rightReleased or justLanded) and not isMoving

local disableKick = leftClicked or rightClicked or justJumped

local isHoldingGun = (activeGun ~= nil and activeGun ~= 0)



-- PLAYER HOTSPOTS MANAGEMENT
-- Don't let hotspot items (arm, mask) move with a latency
local player_vel = EntityGetFirstComponent(player, "VelocityComponent")
local player_vx, player_vy = ComponentGetValueVector2(player_vel, "mVelocity")

local mask = EntityGetWithName("mask")
local mask_x, mask_y = EntityGetTransform(mask)
EntitySetTransform(mask, mask_x + player_vx, mask_y + player_vy)

local e_arm = EntityGetWithName("arm_r")
local arm_x, arm_y = EntityGetTransform(e_arm)
EntitySetTransform(e_arm, arm_x + player_vx, arm_y + player_vy)



-- WEAPON MANAGEMENT
--  * Laser sight show/hide
--  * Weapon throwing enable/disable
if (isHoldingGun) then
  local renderLaser = (
    not holdingUse
    and EntityHasTag(activeGun, "uses_laser_sight")
    and has_perk("LASERSIGHT")
  )
  draw_laser_sight(activeGun, renderLaser)

  local isThrowableGun = EntityHasTag(activeGun, "GUN_THROWN")
  if (useClicked and isThrowableGun) then
    enable_weapon_throwing(activeGun)
  elseif (useReleased and isThrowableGun) then
    disable_weapon_throwing(activeGun)
  end
end



-- OFF-HAND PICKUP MANAGEMENT
--  * Right mouse to pick up any physics items
--  * Disabled when holding a throwable item, because in most cases RMB conflicts with throw key.
if (rmbClicked) then
  -- TODO: Disable when holding a grenade.
  local telekinesis = EntityGetFirstComponentIncludingDisabled(player, "TelekinesisComponent")
  ComponentSetValue2(telekinesis, "mInteract", true)
end



-- KICK MANAGEMENT
--  * If we don't have dash, we kick as usual
--  * With dash, we enable kick only when standing still
if has_perk("DASH") then
  if disableKick then
    enable_kick(player, false)
  elseif enableKick then
    enable_kick(player, true)
  end
end


-- AIR CONTROLS MANAGEMENT
--  * Jump scaling
--  * Double jump
--  * Dashing action
if (inAir) then
  if (not holdingUp) then
    -- Set immediately when player releases the key, even for once.
    STOPPED_SCALING = true
  end

  if (belowScalingThreshold and not STOPPED_SCALING and not HAS_DASHED) then
    jump_scaling(dataComponent, platComponent, framesInAir)
  end

  -- Apply dash after jump_scaling, so it won't interfere
  if (kickClicked and has_perk("DASH")) then
    dash(player, false, dataComponent)
    HAS_DASHED = true
  end

  -- We've stopped holding up, and now only click it once.
  if (holdingUp == true and STOPPED_SCALING and not HAS_DOUBLE_JUMPED and has_perk("DOUBLE_JUMP")) then
    extra_jump(dataComponent)
    HAS_DOUBLE_JUMPED = true
  end

elseif (justLanded) then
  -- Global movement statuses that control the in-air properties.
  -- Must be reset instantly and only upon landing.
  STOPPED_SCALING = false
  HAS_DOUBLE_JUMPED = false
  HAS_DASHED = false

elseif (doGroundDash and has_perk("DASH")) then
  dash(player, true, dataComponent)
  HAS_DASHED = true
end



-- SWIMMING CONTROLS MANAGEMENT
--  * Allow the player to get out of water more easily with our extra movement features
--  * Most important swimming vars are in CharacterPlatformingComponent:
--    - swim_idle_buoyancy_coeff
--    - swim_up_buoyancy_coeff
if (swimmingStarted) then
  HAS_DOUBLE_JUMPED = false
  HAS_DASHED = false
elseif (swimmingStopped) then
  HAS_DOUBLE_JUMPED = false
  HAS_DASHED = false
end



-- Global state handlers, to get custom *clicked*, *released* and other such
-- intermediary triggers. Prevents us from changing components every frame,
-- but only when a change occurs.
PREV_RMB = holdingRmb
PREV_LEFT = holdingLeft
PREV_RIGHT = holdingRight
PREV_KICK = holdingKick
PREV_USE = holdingUse
PREV_IN_AIR = inAir
PREV_FRAMES_IN_AIR = framesInAir
PREV_FRAMES_SWIMMING = framesSwimming
