dofile_once("mods/koira/files/scripts/utilities.lua")
dofile_once("mods/koira/files/scripts/gun_utilities.lua")

-- Manual settings for best results.
AMMO_WARNING_THRESHOLDS = {
  ["SHOTGUNSHELL"] = 2,
  ["PISTOL_CLIP"] = 3,
  ["REVOLVER_AMMO"] = 2,
  ["SMG_CLIP"] = 15,
  ["AR_CLIP"] = 8,
  ["RIFLE_CLIP"] = 2,
  ["CHAINSAW_CANISTER"] = 40,
  ["FMW_CANISTER"] = 50,
  ["ROCKET_MAG"] = 2,
  ["STAKE_BUNDLE"] = 2,
  ["GRENADE_BELT"] = 2,
  ["BOLT_BATTERY"] = 2,
  ["PLASMA_CELL"] = 10,
  ["LIGHTNING_BATTERY"] = 3,
  ["MACHINEGUN_CELL"] = 40,
  ["WATER_CONTAINER"] = 100,
  ["WIND_BLAST"] = 10,
}

-- Cases where we need to update the warnings:
-- * Shooting
-- * Ammo pickup
-- * Ammo drop/throw
-- * Weapon pickup
-- * Weapon drop/throw
-- * Weapon switching
--
-- Thus it's easier (for now) to just constantly update the status,
-- every N frame, according to the currently active item.

local player = GetUpdatedEntityID()
local gun = get_active_gun(player)

local ending_triggered = GlobalsGetValue("koira_ending_triggered", "0") == "1"
if not ending_triggered then
	if (EntityHasTag(gun, "uses_ammo")) then
	  local the_mag = get_active_magazine(gun)

	  if the_mag ~= nil then
		local bullets = get_bullets_left(the_mag)

		-- Action with infinite uses, skip warnings for them.
		if bullets == -1 then return end

		local action_id = get_action_id(the_mag)

		local threshold = AMMO_WARNING_THRESHOLDS[action_id] or 10

		ammo_warning_display(bullets <= threshold)
		mags_warning_display(get_mags_left(gun) <= 1)
		gun_empty_warning_display(false)
	  else
		ammo_warning_display(false)
		mags_warning_display(false)
		gun_empty_warning_display(true)
	  end
	else
	  ammo_warning_display(false)
	  mags_warning_display(false)
	  gun_empty_warning_display(false)
	end
else

  ammo_warning_display(false)
  mags_warning_display(false)
  gun_empty_warning_display(false)

end
